# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module to hold implementations of object service classes."""


import json
import datetime
from typing import BinaryIO, Union, Set, Tuple, Optional

from google.cloud.storage import Client, Bucket, Blob
from google.oauth2.service_account import Credentials


def get_client_credential_content(credential) -> Client:
    """Get a client object for GCS, using environment variable credentials.

    :param credential: The credential content to use.
    """
    credential_data = json.loads(credential)
    creds = Credentials.from_service_account_info(credential_data)
    return Client(credentials=creds)


def get_client(credential_content: Optional[str] = None) -> Client:
    """Get a client object for GCS, using environment variable credentials.

    :param credential_content: The credential content to use, otherwise uses
                                GOOGLE_APPLICATION_DEFAULT.
    :return: A Google storage client instance.
    """
    if credential_content is not None:
        storage_client = get_client_credential_content(credential_content)
    else:
        storage_client = Client()
    return storage_client


def get_bucket(bucket_name: str, client: Client) -> Bucket:
    """Get a bucket object for GCS.

    :param bucket_name: The name of the bucket to get (without the gs://)
    :param client: a Google storage client instance
    :return: A Google storage bucket instance.
    """
    return client.bucket(bucket_name)


def get_bucket_and_client(bucket_name: str,
                          credential_content: Optional[str] = None
) -> Tuple[Bucket, Client]:
    """Get a bucket and client object for GCS.

    :param bucket_name: The name of the bucket to get (without the gs://)
    :param credential_content: The credential content to use, otherwise uses
                                GOOGLE_APPLICATION_DEFAULT.
    :return: tuple of (bucket, client)
    """
    storage_client = get_client(credential_content)
    bucket = get_bucket(bucket_name, storage_client)
    return bucket, storage_client


def validate_url(gs_uri: str,
                 check_exists=None,
                 client: Optional[Client] = None) -> bool:
    """Validate that a given url is a GCS url.

    :param gs_uri: the path to validate
    :param check_exists: flag to check if the path actually exists
    :param client: a Google storage client instance
    """
    if gs_uri.startswith('gs://'):
        if check_exists is True:
            return check_gs_path_exists(gs_uri, client)
        else:
            return True
    else:
        return False


def check_gs_path_exists(gs_uri: str, client: Optional[Client] = None):
    """Check that a provided google cloud storage url refers to an object that exists.

    :param gs_uri: the path to validate
    :param client: a Google storage client instance
    :return: true if an object exists at the url, false otherwise
    """
    bucket_exists = False
    storage_client = Client() if client is None else client

    try:
        bucket_name, file_name, blob_name = split_gs_uri(gs_uri)
    except AttributeError:
        return bucket_exists

    bucket = storage_client.bucket(bucket_name)

    try:
        bucket_exists = bucket.exists()
    except ValueError:
        pass

    if bucket_exists and file_name:
        blob = bucket.get_blob(blob_name)
        return blob is not None
    else:
        return bucket_exists


# TODO: Evaluate the necessity of this method and if it should be removed
# def check_blob_exists(bucket_name, blob_name, client=None):
#     storage_client = Client() if client is None else client
#     bucket = storage_client.bucket(bucket_name)
#     blob = bucket.get_blob(blob_name)
#     return blob.exists()


def split_gs_uri(gs_uri: str) -> Tuple[str, str, str]:
    """Split a gs uri into (bucket, filename, blob).

    :param gs_uri: the uri to split
    """
    try:
        split = gs_uri.split('/')
        return split[2], split[-1], '/'.join(split[3:])
    except IndexError:
        raise AttributeError("Provided url could not be split as a gs url")


# TODO: Evaluate the necessity of this method and if it should be removed
# def download_blob_from_url(client: Client, url: str, dst_file_path: str):
#     with open(dst_file_path) as dst:
#         client.download_blob_to_file(url, dst)


def gs_path_from_blob(blob: Blob) -> str:
    """Determine the full gs URI of a blob.

    :param blob: The blob we want the URI for.
    :return: The blob's full URI.
    """
    return f"gs://{blob.bucket.name}/{blob.name}"


def upload_string_to_bucket_loaction(bucket: Bucket,
                                     data: Union[str, bytes],
                                     content_type: str,
                                     destination_blob_name: str) -> str:
    """Upload a string/bytes obj to a GCS bucket.

    :param bucket: the GCS bucket object to upload the file to
    :param data: the data to upload, something like `response.content`
    :param content_type: the content type, same as `response.headers['content-type']`
    :param destination_blob_name: what to name the file in the bucket
    :return:
    """
    blob = bucket.blob(destination_blob_name)
    blob.upload_from_string(data, content_type=content_type)
    return gs_path_from_blob(blob)


def upload_file_to_bucket_location(bucket: Bucket,
                                   source_file_name: str, destination_blob_name: str) -> str:
    """Upload an io file to a GCS bucket.

    :param bucket: the GCS bucket object to upload the file to
    :param source_file_name: the path location of the io file
    :param destination_blob_name: what to name the file in the bucket
    :return: the GCS path of the uploaded file
    """
    blob = bucket.blob(destination_blob_name)
    blob.upload_from_filename(source_file_name)
    return gs_path_from_blob(blob)


def copy_object_within_gcs(src_bucket: Bucket, src_blob_name: str,
                           dst_bucket: Bucket, dst_blob_name: str, delete=False) -> str:
    """Copy an Object from one GCS location to another.

    :param src_bucket: The bucket containing the source object
    :param src_blob_name: The blob name of the source object
    :param dst_bucket: The bucket to contain the destination object
    :param dst_blob_name: The name of the object to be created in the destination bucket
    :param delete: Flag to delete the source file upon copy completion.
    :return:
    """
    src_blob = src_bucket.blob(src_blob_name)
    dst_blob = src_bucket.copy_blob(src_blob, dst_bucket, dst_blob_name)
    if delete:
        src_bucket.delete_blob(src_blob_name)
    return gs_path_from_blob(dst_blob)


def move_object_within_gcs(src_bucket: Bucket, src_blob_name: str,
                           dst_bucket: Bucket, dst_blob_name: str) -> str:
    """Move an Object from one GCS location to another.

    This is an alias for `copy_object_within_gcs(*args, delete=True)`

    :param src_bucket: The bucket containing the source object
    :param src_blob_name: The blob name of the source object
    :param dst_bucket: The bucket to contain the destination object
    :param dst_blob_name: The name of the object to be created in the destination bucket
    :return:
    """
    return copy_object_within_gcs(src_bucket, src_blob_name,
                                  dst_bucket, dst_blob_name, delete=True)


def generate_signed_url(blob: Blob, expiration=None, ) -> str:
    """Generate a v4 signed URL for uploading a blob using HTTP PUT.

    Note that this method requires a service account key file. You can not use
    this if you are using Application Default Credentials from Google Compute
    Engine or from the Google Cloud SDK.

    See:
     - https://cloud.google.com/storage/docs/access-control/signing-urls-with-helpers#upload-object

    :param blob:
    :param expiration:
    :return:
    """
    expiration = expiration or 24
    return blob.generate_signed_url(
        version="v4",
        expiration=datetime.timedelta(hours=expiration),
        method="GET"
    )


def generate_signed_upload_url(blob: Blob,
                               expiration: Optional[int] = None,
                               content_type: Optional[str] = None):
    """Return a user consumable url for uploading data to the storage object.

    :param expiration: How long the signed url is valid for, in hours.
    :param content_type: The content type of the data to upload, client must set this
    to match.
    :return: A url that the user can use to upload data with a PUT request.
    """
    expiration = expiration or 1
    content_type = content_type or 'application/octet-stream'
    return blob.generate_signed_url(
        version="v4",
        expiration=datetime.timedelta(hours=expiration),
        method="PUT",
        content_type=content_type
    )

# TODO: Evaluate the necessity of this method and if it should be removed
# def download_blob(bucket: Bucket, blob_name, local_file: BinaryIO) -> BinaryIO:
#     bucket.blob(blob_name).download_to_file(local_file)
#     return local_file


def download_blob_by_uri(client: Client,
                         uri: str,
                         local_file: BinaryIO,
                         start: Optional[int] = None,
                         end: Optional[int] = None) -> BinaryIO:
    """Download a blob with nothing but a URI and local file instance.

    :param client: a Google storage client instance
    :param uri: The URI of the Google cloud storage object to download
    :param local_file: The IO to download to
    :param start: the first byte in a range to be downloaded
    :param end: the last byte in a range to be downloaded
    :return: The provided IO after downloading
    """
    client.download_blob_to_file(uri, local_file, start=start, end=end)
    return local_file


def download_as_bytes(client: Client,
                      gs_uri: str,
                      start: Optional[int] = None,
                      end: Optional[int] = None) -> bytes:
    """Given a client and a gs_uri, download and return the contents of the gs_uri as bytes.

    :param client: a Google storage client instance
    :param gs_uri: the absolute gs uri of the object to download
    :param start: the first byte in a range to be downloaded
    :param end: the last byte in a range to be downloaded
    :return: the contents of the object at the gs_uri as bytes
    """
    bucket_name, file_name, blob_name = split_gs_uri(gs_uri)
    bucket = client.bucket(bucket_name)
    blob = bucket.blob(blob_name)
    return blob.download_as_bytes(start=start, end=end)


def all_files(bucket_name,
              prefix=None,
              full_gs_path=True,
              client: Optional[Client] = None) -> Set[str]:
    """List all files at the provided location.

    :param bucket_name: Name of the bucket
    :param prefix: Prefix to apply to the objects
    :param full_gs_path: The full path to search for.
    :return: A set of all the objects at the location.
    :param client: a Google storage client instance
    """
    storage_client = Client() if client is None else client
    if full_gs_path:
        return {f"gs://{bucket_name}/{blob.name}" for blob in
                storage_client.list_blobs(bucket_name, prefix=prefix)}
    else:
        return {blob.name for blob in
                storage_client.list_blobs(bucket_name, prefix=prefix)}


def delete_blob(blob: Blob):
    """Delete a blob from GCS.

    :param blob: The blob to delete.
    """
    blob.delete()


# TODO: Evaluate the necessity of this method and if it should be removed
# def find_subfolder(bucket_name, prefix=None):
#     folders = set()
#
#     storage_client = Client()
#     blobs = storage_client.list_blobs(bucket_name, prefix=prefix)
#     for blob in blobs:
#         # print(blob.name)
#         name = blob.name
#         parts = name.split('/')
#         folder_name = parts[0] if prefix is None else parts[1]
#         folders.add(folder_name)
#
#     return folders


# TODO: Evaluate the necessity of this method and if it should be removed
# def find_file(bucket_name, metadata_file_name, prefix=None):
#     metadata_files = []
#
#     storage_client = Client()
#     blobs = storage_client.list_blobs(bucket_name, prefix=prefix)
#     prefix_len = len(prefix)
#     for blob in blobs:
#         name = blob.name
#         if name.endswith(metadata_file_name):
#             if prefix is None:
#                 metadata_files.append(name)
#             else:
#                 metadata_files.append(name[prefix_len:])
#     return metadata_files


# TODO: Evaluate the necessity of this method and if it should be removed
# def find_file_startswith(bucket_name, file_name_startswith, bucket_folder_prefix=None):
#     files = []
#
#     storage_client = Client()
#     blobs = storage_client.list_blobs(bucket_name, prefix=bucket_folder_prefix)
#     prefix_len = len(bucket_folder_prefix)
#     for blob in blobs:
#         name = blob.name
#         if name.endswith("/"):
#             continue
#         parts = name.split("/")
#         last = parts[-1]
#
#         if last and last.startswith(file_name_startswith):
#             files.append(name)
#     return files



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A namespace of exceptions used by this package.

WARNING: This is likely to change as this interface redesigned and refactored!

TODO: Redesign and refactor this file to better reflect how exceptions should be
    handled throughout the rest of the package. Make sure to document the ABCs
    as to how they expect their implementations to handle exceptions.
"""


class BaseError(BaseException):
    """The root of all exceptions in this package."""


class ApiError(BaseError):
    """Used for when the package runs into an error connecting to another service."""


class NotFoundError(ApiError):
    """For when a referenced object can't be found."""


class ApiAuthError(BaseError):
    """For when a referenced object can't be accessed due to auth problems."""



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The abstract class definitions for ObjectSvc and ObjectIngestSvc implementations."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Type, Dict, BinaryIO, ClassVar, Optional, Set

from jax.cs.storage.config import StorageConfig, default_settings
from jax.cs.storage.enums import UrlTypes


class ObjectSvc(ABC):
    """The base abstract class for all storage service backends.

    A ObjectSvc implementation should provide these methods to a specific file backend
    (e.g. gcs, s3, ftp). This allows the StorageObject to use a consistent interface for file
    interaction with the service. A concrete implementation of this abstract class can be used
    to configure the StorageObject to handle files of that specific file backend.
    """

    CONFIG: ClassVar[StorageConfig] = default_settings

    def __init__(self, url: str, content_type: Optional[str] = None, **kwargs):
        """Create an instance of an ObjectSvc.

        The url represents the location of the storage object.
        :param url: The url of the object the service will handle
        :param content_type: The content type of the file
        :param kwargs: keyword arguments to be used when handling the file
        """
        self.url = url
        self.content_type = content_type or 'text/plain'
        self.kwargs = kwargs

    @classmethod
    @abstractmethod
    def validate(cls, url: str) -> bool:
        """Validate that the service handle the provided url.

        :return: True: the url can be handled, False: the url cannot be handled
        """

    @classmethod
    @abstractmethod
    def all(cls, root: Optional[str] = None, **kwargs) -> Set[str]:
        """Return a list of urls representing files in the service.

        :param root: A base directory to list from
        :return: A set of all urls that the service can determine it contains
        """

    @classmethod
    @abstractmethod
    def exists(cls, sub_location: str, root: Optional[str] = None) -> Optional[str]:
        """Check if an object exists.

        Given a filename/path within the service, return the full uri if it exists or None otherwise
        :param sub_location: The path of the file without the scoping of the service
                                e.g. 'blob_path.txt' instead of 'gs://bucket/blob_path.txt'
                                       or
                                     'subdir/myfile.txt' instead of '/basedir/subdir/myfile.txt'
        :param root: Convenience arg to search only below a specific root.
                                e.g. sub_location = 'blob_path.txt' and root = 'my_gcs_dir' then
                                    the method will search for 'my_gcs_dirt/blob_path.txt' within
                                    the backing file service and return the fully qualified path if
                                    found.
        :return: full uri (e.g. 'gs://bucket/blob_path.txt') if it exists or None otherwise
        """

    @abstractmethod
    def available(self) -> Dict[UrlTypes, ObjectSvc]:
        """Return the available ways that the service can provide the file.

        A method that returns the available ways that the underlying ObjectSvc can provide the file.
        :return: A dictionary mapping UrlTypes to ObjectSvc instances
        """

    @abstractmethod
    def download(self,
                 local_file: BinaryIO,
                 start: Optional[int] = None,
                 end: Optional[int] = None) -> BinaryIO:
        """Download an instance of a storage object to a provided BinaryIO instance.

        Each service should download the file according to the implemented downloader and return
        the new location of the downloaded file.

        Some services might support downloading only part of the object, in which case they should
        use the optional `start` and `end` parameters to determine which part of the object to
        download.

        :param local_file: a file like object to download to
        :param start: (if supported) the first byte in a range to be downloaded
        :param end: (if supported) the last byte in a range to be downloaded
        :return: The io file with content written to it
        """

    @abstractmethod
    def download_as_bytes(self,
                          start: Optional[int] = None,
                          end: Optional[int] = None) -> bytes:
        """Retrieve the contents of the file as bytes.

        Some services might support downloading only part of the object, in which case they should
        use the optional `start` and `end` parameters to determine which part of the object to
        download.

        :param start: (if supported) the first byte in a range to be downloaded
        :param end: (if supported) the last byte in a range to be downloaded
        :return: The file as bytes
        """

    def replace(self, replacement: ObjectSvc):
        """Replace the content of the storage object with another storage object.

        Replaces the underlying data represented by this storage object svc instance
        with the data from another storage object svc instance.

        :param replacement: The replacement storage object.
        """

    @abstractmethod
    def delete(self):
        """Delete the underlying data represented by this storage object instance."""

    @property
    @abstractmethod
    def user_url(self) -> str:
        """Return a user consumable url for downloading the storage object.

        :return: A url representation of the file location for the user
        """

    def user_upload_url(self, **kwargs) -> str:
        """Return a user consumable url for uploading data to the storage object.

        :return: A url that the user can use to upload data with a PUT request.
        """

    @property
    @abstractmethod
    def info(self) -> Dict[str, str]:
        """Information about the storage object.

        :return: A dictionary of all information the service can determine about the storage object
        """

    @abstractmethod
    def refresh_info(self) -> Dict[str, str]:
        """Refresh the information about the storage object.

        :return: A dictionary of all information the service can determine about the storage object
        """


class ObjectIngestSvc(ABC):
    """An ObjectSvc that ingests and builds instances of ObjectSvc.

    A special factory class for ObjectSvc instances that handles moving files into a specific
    ObjectSvc backend using the `.ingest()` method.
    """

    BASE_SVC: Type[ObjectSvc]

    def __init__(self, base_svc: Optional[Type[ObjectSvc]] = None):
        """Construct an ingestion service class.

        The constructor takes a base ObjectSvc which can be passed by the end-user, or dynamically
        with a call to super().

        :param base_svc: Optionally override the base service (note: it is up to the developer to
        determine that the ObjectSvc implementation works with the ObjectIngestSvc implementation).
        """
        self._base_svc: Type[ObjectSvc]

        if base_svc is not None:
            self._base_svc = base_svc

    @property
    def base_svc(self) -> Type[ObjectSvc]:
        """Return the service class of which an instance will be built when ingest is called.

        :return: The ObjectSvc class the file will be moved to
        """
        return self._base_svc

    @abstractmethod
    def ingest(self, ext_file: ObjectSvc, **kwargs) -> ObjectSvc:
        """Build an ObjectSvc instance from the provided ObjectSvc arg and return it.

        A factory method to build a ObjectSvc representing a file that is internal to the system
        from a ObjectSvc instance representing a file external to the system.

        :param ext_file:
        :param kwargs: keyword args to be used by the ingestion service
        :return: An instance of a ObjectSvc after representing the file after being moved
        """

    def generate_internal_name(self,
                               filename: Optional[str] = None,
                               **kwargs) -> str:
        """Format the name of the ingested object.

        A template method to override generation of internal file names

        :param filename: The base filename prior to ingestion of the file
        :param kwargs: To be used to pull in extra information to be used in the naming process
        :return: The internal name of the file to use
        """



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper init function to make it easy to configure package settings."""

from typing import List, Type, Tuple, Optional

from jax.cs.storage.base import StorageObject, StorageObjectIngest
from jax.cs.storage.config import StorageConfig
from jax.cs.storage.object.abstract import ObjectSvc, ObjectIngestSvc

init_return = Tuple[Type[StorageObject], Type[StorageObjectIngest]]


def init(settings: Optional[StorageConfig] = None,
         default_svc: Optional[Type[ObjectSvc]] = None,
         services: Optional[List[Type[ObjectSvc]]] = None,
         ingestion_svc: Optional[Type[ObjectIngestSvc]] = None) -> init_return:
    """Build StorageObject, StorageObjectIngest by helping with configuration and customization.

    :param settings: explicit settings to use with the library,
                        otherwise uses .env/environment variables
    :param default_svc: an ObjectSvc to use as the default service
    :param services: a list of ObjectSvc to select from when creating a StorageObject
    :param ingestion_svc: an ObjectIngestSvc to use to ingest storage objects
    :return: an un-instantiated StorageObject and StorageObjectIngest
    """
    if settings:
        ObjectSvc.CONFIG = settings

    if services:
        StorageObject.SVCS = services

    if default_svc:
        StorageObject.DEFAULT_SVC = default_svc

    if ingestion_svc:
        StorageObjectIngest.INGESTION_SERVICE = ingestion_svc

    return StorageObject, StorageObjectIngest



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The root of the jax.cs.storage module.

e.g.

from jax.cs.storage import StorageObject, StorageObjectIngest

OR

from jax.cs.storage.entrypoint import init
StorageObject, StorageObjectIngest = init()
"""
try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata

__version__ = metadata.version("jax-cs-storage")

from jax.cs.storage.base import StorageObjectIngest, StorageObject
from jax.cs.storage.entrypoint import init
from jax.cs.storage.enums import UrlTypes
from jax.cs.storage.object.services.gcs import ObjectSvcGCS, ObjectIngestSvcGCS
from jax.cs.storage.object.services.io import ObjectSvcIO
from jax.cs.storage.object.services.r2 import ObjectSvcR2

__all__ = ['StorageObject', 'StorageObjectIngest', 'init', 'UrlTypes', 'ObjectSvcGCS',
           'ObjectIngestSvcGCS', 'ObjectSvcIO', 'ObjectSvcR2', '__version__']



"""A default object cache implementation."""

import os
import pickle
import typing


from jax.cs.storage import init
from jax.cs.storage.object.services.gcs import ObjectIngestSvcGCS
from jax.cs.storage.object.services.io import ObjectIngestSvcIO
from jax.cs.storage.config import StorageConfig


class ObjectCache:
    """Class for saving and loading objects from a cache."""

    def __init__(self, use_gcs: bool = True,
                 wrapper_config: typing.Optional[typing.Dict] = None):
        """Initialize a new or existing cache.

        :param use_gcs: if True, use Google Cloud Storage backend
        :param wrapper_config: optional parameter to configure the
        jax-cs-filewrapper backend used for storing the cache files. By
        default, the configuration is set through a .env file or environment
        variables.
        """
        self._use_gcs = use_gcs

        # configure the default ingestion service for the filewrapper module
        ingestion_svc = ObjectIngestSvcGCS if use_gcs else ObjectIngestSvcIO

        if wrapper_config is None:
            # get the configuration from the environment
            self._FileWrapper, self._FileIngester = init(
                ingestion_svc=ingestion_svc
            )
        else:
            # optional dict with configuration values for jax.cs.filewrapper
            # has been passed for has been passed in. use that to manually
            # configure file wrapper
            self._FileWrapper, self._FileIngester = init(
                StorageConfig.parse_obj(wrapper_config),
                ingestion_svc=ingestion_svc
            )

        self._BackendFileSvc = self._FileIngester.resulting_svc()

        # if using local, make sure the directory exists
        if use_gcs:
            if self._BackendFileSvc.CONFIG.GCS_BUCKET is None:
                raise RuntimeError("JAX_CS_STORAGE_GCS_BUCKET "
                                   "environment variable not defined")

        elif not os.path.exists(self._BackendFileSvc.CONFIG.IO_FILE_ROOT):
            raise RuntimeError("Local object cache directory does not exist")

    def save(self, obj: object, filename: str) -> None:
        """Save an object to the cache.

        :param obj: Python object to save (must be 'pickleable')
        :param filename: filename to save the object to in the cache
        :return: None
        """
        self._FileIngester(pickle.dumps(obj), from_memory=True).ingest(
            filename=filename, content_type='application/octet-stream')

    def load(self, filename: str) -> typing.Any:
        """Load an object from the cache.

        :param filename: filename for the object in the cache
        :return: Python object
        :raises: FileNotFound error if filename does not exist in cache
        """
        path = self._BackendFileSvc.exists(filename)
        if path is None:
            raise FileNotFoundError

        wrapped = self._FileWrapper(path)
        return pickle.loads(wrapped.download_as_bytes())

    def delete(self, filename: str) -> None:
        """Delete a file from the cache.

        :param filename: filename to delete
        :return: None
        """
        # TODO at the time of development, jax-cs-filewrapper hasn't implemented
        #  the delete method
        # delete not yet implemented in cs-file-wrapper
        pass

    def check_if_exists(self, filename: str) -> bool:
        """Check if a file exists in the object cache.

        :param filename: filename of cached object
        :return: true of file found in cache, false otherwise
        """
        return self._BackendFileSvc.exists(filename) is not None

    def cache_listing(self) -> typing.Set[str]:
        """List objects in the cache.

        :return: The objects in the cache.
        """
        listing = self._BackendFileSvc.all()
        if not self._use_gcs:
            # for local files we want to remove the fully qualified path and
            # have just the files relative to the root directory
            root_dir = self._BackendFileSvc.CONFIG.IO_FILE_ROOT
            if not root_dir.endswith(os.path.sep):
                root_dir = root_dir + os.path.sep

            listing = {
                f.replace(root_dir, '') for f in listing
            }
        return listing



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Storage Object Services backed by Jax/CS's R2 web file service."""

import logging
from typing import BinaryIO, Dict, Optional, Set

from jax.cs.storage.object.abstract import ObjectSvc, UrlTypes
from jax.cs.storage.object.services.gcs.service import ObjectSvcGCS
from . import utils


class ObjectSvcR2(ObjectSvc):
    """An object storage service backed by Jax/CS's R2 web file service."""

    def __init__(self, url: str, **kwargs):
        """Create an instance of the service.

        :param url: The url of the file in r2.
        :param kwargs: If passed, the `auth_header` and `r2_url` kwargs will be used.
        """
        logging.warning(
            "R2 service is deprecated and will be removed in a future release."
        )
        super().__init__(url, **kwargs)
        self.auth_header = kwargs.get('auth_header')
        self.r2_url = kwargs.get('r2_url')
        self.__info_cache = None
        if not self.auth_header:
            raise AttributeError("Auth header required for r2 service")

    @classmethod
    def all(cls, root: Optional[str] = None, **kwargs) -> Set[str]:
        raise NotImplementedError

    @classmethod
    def validate(cls, url: str) -> bool:
        return utils.validate_url(url, r2_base_url=cls.CONFIG.R2_BASE_URL)

    @classmethod
    def exists(cls, sub_location: str, root: Optional[str] = None) -> Optional[str]:
        sub_location = f"{root}/{sub_location}" if root else sub_location
        full_url = f"{cls.CONFIG.R2_BASE_URL}/{sub_location}"
        return full_url if cls.validate(full_url) else None

    def available(self) -> Dict[UrlTypes, ObjectSvc]:
        return {
            UrlTypes.R2: self,
            UrlTypes.GCS: ObjectSvcGCS(self._gcs_url)
        }

    def download(self,
                 local_file: BinaryIO,
                 start: Optional[int] = None,
                 end: Optional[int] = None) -> BinaryIO:
        return utils.download_file(self.user_url, local_file)

    def download_as_bytes(self,
                          start: Optional[int] = None,
                          end: Optional[int] = None) -> bytes:
        return NotImplemented

    def delete(self):
        raise NotImplementedError

    @property
    def info(self):
        if not self.__info_cache:
            self.__info_cache = utils.get_url_json_or_error(self.url, self.auth_header)
        return self.__info_cache

    @property
    def user_url(self) -> str:
        return self.info.get('location')

    def user_upload_url(self, **kwargs) -> str:
        raise NotImplementedError

    @property
    def _gcs_url(self) -> str:
        return 'gs://' + '/'.join(self.user_url.split('?')[0].split('/')[3:])

    def refresh_info(self) -> Dict[str, str]:
        self.__info_cache = utils.get_url_json_or_error(self.url, self.auth_header)
        return self.__info_cache



import os
import uuid
from contextlib import contextmanager
from typing import Type, Optional, Sequence, BinaryIO, List

from .abstract import FileSvc

class NoOpDownloader(FileSvc):
    """
    An extension of Downloader class that matches all files. Used as the default downloader for
    files that match no other configured downlaoder.
    """

    @classmethod
    def validate(cls, location: str):
        """
        All files are valid for the NoOpDownloader.
        :param location: the uri/url of the file to validate
        """
        return True


class Localizer:
    """
    A class to help organize the downloads of multiple types files in a single arbitrarily nested
    python datasetructure.
    """

    FILE_JOIN_CHAR = '_'

    def __init__(self,
                 root: str,
                 downloaders: Sequence[Type[FileSvc]],
                 no_match: Optional[Type[FileSvc]] = None,
                 **kwargs
                 ):
        self.root = root
        self.downloaders = downloaders
        self.kwargs = kwargs
        self.no_match_downloader = NoOpDownloader if no_match is None else no_match

    def localize_json_dict(self, inputs: dict):
        """
        Given a python dictionary representing a json object, iterate through it to localize each
        file that it references.
        """
        for k, v in inputs.items():
            inputs[k] = self.smart_process(k, v)
        return inputs

    def smart_process(self, key, value, keychain: Sequence = None):
        """
        Given a value that's either a list, a dict, a string, or something else, determine what
        should be done with it. All values are processed further except for "something else" which
        is returned as is.
        """
        keychain = [] if keychain is None else keychain
        if type(value) is list:
            return [i for i in self.process_list(value, keychain)]
        elif type(value) is dict:
            return {k: v for k, v in self.process_dict(value, keychain)}
        elif type(value) is str:
            return self.process_str(key, keychain)
        else:
            return key, value

    def process_list(self, item: list, keychain: List):
        """
        Given a list with nested input files to download, iterate over the list
        and process the items within it
        :param keychain:
        :param item: a python list with nested input files to be downloaded
        """
        for elem in item:
            yield self.smart_process(elem, keychain + ['1'])

    def process_dict(self, item: dict, keychain: List):
        """
        Given a dict with nested input files to download, iterate over the dict
        and process the items within it
        :param keychain:
        :param item: a python dict with nested input files to be downloaded
        """
        for k, v in item.items():
            yield k, self.smart_process(v, keychain + [k])

    def process_str(self, item: str, keychain: List):
        """
        Given a string representation of an item location, download it
        :param keychain:
        :param item: The string location of the item to process
        """
        d = self.select_downloader(item)(item, **self.kwargs)
        # return d.download(self.__get_file_io())

    def select_downloader(self, item: str) -> Type[FileSvc]:
        """
        Given a uri/url item location, select the first downloader defined
        in this instance that validates the item.
        :param item: The string location of the item to select downloader for
        """
        downloader = self.no_match_downloader
        for d in self.downloaders:
            if d.validate(item):
                downloader = d
                break

        return downloader

    def download(self, item: str):
        """
        An alias for `process_str`
        :param item: The string location of the item to download
        """
        return self.process_str(item)

    # @classmethod
    # def generate_local_name(cls, keychain: List[str], file_info: Optional[dict] = None):
    #     kc = keychain.copy()
    #     try:
    #         kc.append(file_info['name'])
    #     except KeyError:
    #         kc.append(uuid.uuid4().hex)
    #     f_name = cls.FILE_JOIN_CHAR.join(keychain)
    #     r_name = file_info.get('name')
    #     if r_name:
    #         f_name =
    #     return os.path.join(self.root, )
    #     ...

    # @contextmanager
    # def file_io(self, file_path=None) -> BinaryIO:
    #     file_path = self.__generate_local_name() if file_path is None else file_path
    #     try:
    #         yield ...
    #     finally:
    #
    #     ...



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Storage Object Services backed by Google Cloud Storage."""

import logging
import shutil
import requests
from enum import Enum
from typing import Type, BinaryIO
from urllib.parse import urlparse


from jax.cs.storage.exceptions import ApiAuthError


def get_url_json_or_error(url, auth_header):
    """Return the json from a GET on a url, ohterwise raise a standard error.

    A helper method to return json of a response if the status is not in the 4xx or 5xx range.
    :param url:  The url to get
    :param auth_header: The auth header to use
    :raises ApiAuthError: if the response status != 200
    :return: The json of the response
    """
    resp = requests.get(url, headers=auth_header)
    if resp.ok:
        return resp.json()
    raise ApiAuthError(f'R2 returned status [{resp.status_code}] on GET: {resp.content}')


def get_file_from_fileset_by_name(file_set, name_contains: str, auth_header: dict):
    """Get the first file in a file_set where `name_contains` is in the file's name.

    :param file_set: The array of file urls
    :param name_contains: The string the file name should contain
    :param auth_header: The auth header to use
    :return: file url if found, otherwise None
    """
    file_url = None
    for file_url in file_set:
        resp = get_url_json_or_error(file_url, auth_header)
        if name_contains in resp.get('name', ''):
            file_url = file_url
            break
    return file_url


def get_file_from_data_object_by_name(dataobject_url, name_contains: str, auth_header: dict):
    """Get the first file in a data_object where `name_contains` is in the file's name.

    :param dataobject_url: The array of file urls
    :param name_contains: The string the file name should contain
    :param auth_header: The auth header to use
    :return: file url if found, otherwise None
    """
    resp = get_url_json_or_error(dataobject_url, auth_header)
    file_url = get_file_from_fileset_by_name(resp.get('file_set', []),
                                             name_contains=name_contains,
                                             auth_header=auth_header)
    if not file_url:
        logging.error(f"Data object has no file with {name_contains} in the name")
        raise ApiAuthError(f"Data object has no file with {name_contains} in the name")

    return file_url


# NOTE: name_enum is of type Type[Enum] to specify that the input is a subclass of Enum
# - https://www.python.org/dev/peps/pep-0484/#the-type-of-class-objects
def get_files_from_fileset_by_name(file_set, name_enum: Type[Enum], auth_header: dict):
    """Get all files in a file_set where one of `name_enum.values()` is in the file's name.

    Quits when it's looked at all the files, or has a file for each enum.
    :param file_set: The array of file urls
    :param name_enum: An enum whose values are used to match the name, the
    matched enum is returned in a tuple with along with the file_url
    :param auth_header: The auth header to use
    :return: list of tuples of the file url and enum type, if found, otherwise None
    """
    file_urls = []
    enum_values = [v.value for v in name_enum.__members__.values()]

    for file_url in file_set:
        resp_dict = get_url_json_or_error(file_url, auth_header)
        name = resp_dict['name']
        for v in enum_values:
            if v in name:
                file_urls.append((file_url, v))
                break
        if len(file_urls) == len(name_enum):
            break

    return file_urls


def get_files_from_data_object_by_name(dataobject_url, name_enum: Type[Enum], auth_header: dict):
    """Get files from a data object.

    A convenience method to get files from a data object, will call `get_url_json_or_error` and
    `get_files_from_fileset_by_name`.
    :param dataobject_url: The url of the data object with the file_set of files to get
    :param name_enum: An enum whose values are used to match the name, the
    matched enum is returned in a tuple with along with the file_url
    :param auth_header: The auth header to use
    :return: list of tuples of the file url and enum type, if found, otherwise None
    """
    resp = get_url_json_or_error(dataobject_url, auth_header)
    return get_files_from_fileset_by_name(resp.get('file_set', []),
                                          name_enum=name_enum,
                                          auth_header=auth_header)


def validate_url(url, r2_base_url):
    """Validate that a string is a url.

    :param url: Url to validate
    :return: bool representing if the string was a url
    """
    result = urlparse(url)
    netloc = urlparse(r2_base_url).netloc
    return all(
        [result.scheme, result.scheme == 'https', result.netloc == netloc])


def download_file(https_url: str, local_file: BinaryIO):
    """Download the contents of an https location to an io instance.

    :param https_url: The url location of the file to download
    :param local_file: The io instance to download to
    :return: The provided io instance
    """
    logging.info(f"Downloading {https_url}")
    with requests.get(https_url, stream=True) as request:
        shutil.copyfileobj(request.raw, local_file)
    return local_file



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility methods used by this package internally."""

import importlib
import logging

from types import ModuleType
from typing import Tuple, List, Optional

# TODO: Remove when reviewed for external package use
# from typing import List
# def try_to_add_import(to_add_to: List, module_name: str):
#     import importlib
#     try:
#         mod = importlib.import_module(module_name)
#         to_add_to.append(mod)
#     except ImportError:
#         pass
#     return to_add_to


def opt_get_module(module_name) -> Optional[ModuleType]:
    """Silently fail on module imports.

    :param module_name: The name of the module to import.
    :return: The module, if imported otherwise None.
    """
    try:
        return importlib.import_module(module_name)
    except ImportError as e:
        logging.error(e)
        return None


def get_available_svcs(services: List[Tuple[str, str]]) -> List:
    """Import and return only the items whose import doesn't cause an ImportError or AttributeError.

    :param services: The first item in the tuple is the python import style module location, and
    the second is the name of the item to import from that module.
    :return: A list of the services that could be imported from their corresponding module.
    """
    ret = []
    for svc in services:
        mod_name, svc_name = svc
        try:
            mod = opt_get_module(mod_name)
            ret.append(getattr(mod, svc_name))
        except (ImportError, AttributeError):
            logging.warning(f"jax.cs.storage could not import service: {svc_name}")
            continue
    return ret



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The base namespace for interacting with this package.

Provides wrapper and factory classes to provide an abstraction to the underlying ObjectSvc
implementations.
"""

import logging
from typing import Type, Dict, BinaryIO, Optional, List, Union
from google.auth.exceptions import DefaultCredentialsError

from jax.cs.storage.object.abstract import ObjectSvc, UrlTypes, ObjectIngestSvc
from jax.cs.storage.object.services.gcs import ObjectIngestSvcGCS
from jax.cs.storage.object.services.io import ObjectSvcInMemory
from jax.cs.storage.utils import get_available_svcs

__all__ = ['StorageObject', 'StorageObjectIngest']


class StorageObject:
    """A wrapper around storage services to make various storage services look the same.

    A Facade for ObjectSvc classes that automates the selection and creation of the concrete
    ObjectSvc implementations.
    """

    SVCS = get_available_svcs(
        [
            ('jax.cs.storage.object.services.gcs', 'ObjectSvcGCS'),
            ('jax.cs.storage.object.services.io', 'ObjectSvcIO'),
        ]
    )
    DEFAULT_SVC = SVCS[0]

    def __init__(self,
                 url: str,
                 file_svc: Optional[Union[ObjectSvc, Type[ObjectSvc]]] = None,
                 file_svcs: Optional[List[Type[ObjectSvc]]] = None,
                 content_type: Optional[str] = None,
                 **kwargs):
        """Create a wrapped storage object with the storage object wrapper.

        :param url: The url of the storage object to wrap
        :param file_svc: manually override service selection by providing an instantiated service
        :param file_svcs: manually override priority or service selection by providing a
        list of uninstantiated ObjectSvc classes to use ordered by priority.
        :param content_type: The content type of the file (optional)
        :param kwargs: All kwargs will be passed to underlying storage object service creation
        """
        self._url = url
        self._kwargs = kwargs
        self._svc_list = file_svcs or self.SVCS
        _svc = file_svc or self.get_svc(url, self._svc_list)

        # This bit of code checks to see if the svc is instantiated or not. If it is not,
        # the default case, then it is instantiated here. Otherwise, it is used as is.
        if isinstance(_svc, type) and issubclass(_svc, ObjectSvc):
            self.svc: ObjectSvc = _svc(url, content_type=content_type, **kwargs)
        elif not isinstance(_svc, type) and isinstance(_svc, ObjectSvc):
            self.svc = _svc
        else:
            raise AttributeError("Provided file_svc must be either a subclass of "
                                 "ObjectSvc or an instance of a subclass of ObjectSvc")

    @classmethod
    def get_svc(cls,
                url: str,
                svc_list: Optional[List[Type[ObjectSvc]]] = None,
                default_svc: Optional[Type[ObjectSvc]] = None) -> Type[ObjectSvc]:
        """Select a suitable storage object service.

         Checks the url against each services' `.validate` method.

        :param url: The url of the storage object to select a service for
        :param svc_list: manually override priority or service selection by providing a list of
        uninstantiated ObjectSvc classes to use ordered by priority.
        :param default_svc: manually override  default service selection by providing an
        uninstantiated service
        :return: The selected uninstantiated service
        """
        found = False
        svc = default_svc or cls.DEFAULT_SVC
        svcs = svc_list or cls.SVCS
        for s in svcs:
            try:
                if s.validate(url):
                    svc = s
                    found = True
                    break
            except (OSError, DefaultCredentialsError) as e:
                logging.warning(f"Unable to validate {s.__name__} for {url}: {e}")
                continue
        if not found:
            logging.error(f"No suitable file service found, defaulting to {cls.DEFAULT_SVC}")
        return svc

    @classmethod
    def validate(cls,
                 url: str,
                 svc_list: Optional[List[Type[ObjectSvc]]] = None,
                 default_svc: Optional[Type[ObjectSvc]] = None) -> bool:
        """Test if the service handle the provided url.

        :param url: The url to validate
        :param svc_list: An optional alternate list of ObjectSvc implementations to validate against
        :param default_svc: An optional alternate default service to use when no others match
        :return: A boolean value representing if at least one of the available services can handle
        the provided url
        """
        return cls.get_svc(url, svc_list, default_svc).validate(url)

    def available(self) -> Dict[UrlTypes, str]:
        """Return the available ways that the underlying ObjectSvc can provide the file."""
        return self.format_available_hook(self.svc.available())

    def format_available_hook(self,
                              available_items: Dict[UrlTypes, ObjectSvc]) -> Dict[UrlTypes, str]:
        """Templated hook method to allow a developer to customize the availability dictionary.

        :param available_items: The result of calling the `available` method on the underlying
        ObjectSvc implementation.
        :return: a mapping from jax.cs.storage.enum.UrlTypes a string representation of the
        storage object
        """
        return {k: v.url for k, v in available_items.items()}

    def download(self, local_file: BinaryIO) -> BinaryIO:
        """Download an instance of a storage object to a provided BinaryIO instance.

        Each service should download the file according to the implemented downloader and return the
         new location of the downloaded file.
        :param local_file: a file like object to download to
        """
        return self.svc.download(local_file)

    def download_as_bytes(self) -> bytes:
        """Download and return the storage object as bytes.

        :return: The contents of the storage object as bytes
        """
        return self.svc.download_as_bytes()

    def delete(self):
        """Delete the underlying data represented by this storage object instance."""
        self.svc.delete()

    def replace(self, replacement: Union[str, 'StorageObject']):
        """Replace the underlying data represented by this storage object instance."""
        if isinstance(replacement, str):
            replacement = StorageObject(replacement)

        self.svc.replace(replacement.svc)

    @property
    def url(self) -> str:
        """Return the full url representation of the storage object.

        :return: The full url
        """
        return self.svc.url

    @property
    def user_url(self) -> str:
        """Return a user consumable url for the storage object.

        :return: User consumable url for the storage object
        """
        return self.svc.user_url

    def user_upload_url(self, **kwargs) -> str:
        """Return a user consumable url for uploading data to the storage object.

        :return: A url that the user can use to upload data with a PUT request.
        """
        return self.svc.user_upload_url(**kwargs)

    @property
    def user_name(self) -> str:
        """Return a user consumable name for the storage object.

        :return: User consumable name for the storage object
        """
        name = self.info.get('name')
        return name.split('/')[-1] if name else 'NO_NAME'

    @property
    def info(self) -> Dict[str, str]:
        """Information about the storage object.

        A proxy method for the underlying ObjectSvc .info method.

        :return: A dictionary of all information the underlying service knows about the storage
        object
        """
        return self.svc.info

    def refresh_info(self) -> Dict[str, str]:
        """Refresh the information about the storage object.

        A proxy method for the underlying ObjectSvc .refresh_info method.

        :return: A dictionary of all information the underlying service knows about the storage
        object
        """
        return self.svc.refresh_info()


class StorageObjectIngest(StorageObject):
    """A default implementation of a File Ingestion Facade.

    This Facade extends the StorageObject facade to include a factory method for StorageObject
    instances that need to be imported/ingested into the system. Note that this class inherits from
    StorageObject, and thatit is fully usable as astorage object wrapper on the remote storage
    object.

    Non-system File -> Ingester(Non-system File) -> Wrapper(System File)
    """

    INGESTION_SERVICE: Type[ObjectIngestSvc] = ObjectIngestSvcGCS

    def __init__(self,
                 url: str,
                 file_svc: Optional[ObjectSvc] = None,
                 file_svcs: Optional[List[Type[ObjectSvc]]] = None,
                 from_memory: bool = False,
                 **kwargs):
        """Create and instance of a storage object ingestion class.

        Extension of the `jax.cs.storage.base.StorageObject` constructor to include a `from_memory`
        kwarg flag. This flag will treat the url as the object contents rather than the object
        location, which is useful to create new files from content generated within an application.

        :param url: The url of the storage object to wrap
        :param file_svc: manually override service selection by providing an instantiated service
        :param file_svcs: manually override priority or service selection by providing a list of
        uninstantiated ObjectSvc classes to use ordered by priority.
        :param from_memory: If the url provided should be treated as memory. If so, the content of
         the provided url argument will be used as the content of the new ingested file.
        :param kwargs: All kwargs will be passed to underlying storage object service creation
        """
        if from_memory:
            file_svc = ObjectSvcInMemory(url=url, **kwargs)
        super().__init__(url, file_svc=file_svc, file_svcs=file_svcs, **kwargs)

    def ingest(self, **kwargs) -> StorageObject:
        """Create a StorageObject instance from ingesting the provided object.

        This method is a Factory for StorageObject instances.

        The ingest method of a ObjectIngestSvc is a factory for ObjectSvc instances.
        """
        self.svc = self._call_ingestion_svc(**kwargs)
        return self._create_new_file_wrapper()

    def _call_ingestion_svc(self, **kwargs) -> ObjectSvc:
        """Intended as a template method that can be overridden if needed."""
        return self.INGESTION_SERVICE().ingest(self.svc, **{**self._kwargs, **kwargs})

    def _create_new_file_wrapper(self, **kwargs):
        """Intended as a template method that can be overridden if needed."""
        return StorageObject(self.svc.url, file_svc=self.svc, **kwargs)

    @classmethod
    def resulting_svc(cls) -> Type[ObjectSvc]:
        """Return the uninstantiated class that the `ingest()` method will return an instance of."""
        return cls.INGESTION_SERVICE().base_svc



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Storage object service backed by python IO."""

import gc
import logging
import os
import shutil
import uuid
from typing import Dict, BinaryIO, Set, Optional, Type

from jax.cs.storage.object.abstract import ObjectSvc, ObjectIngestSvc, UrlTypes


class ObjectSvcIO(ObjectSvc):
    """A service for interacting with file like IO as objects."""

    def __init__(self, url: str, **kwargs):
        super().__init__(url, **kwargs)
        self._info_cache = None

    @classmethod
    def exists(cls, sub_location: str, root: Optional[str] = None) -> Optional[str]:
        sub_location = os.path.join(root, sub_location) if root else sub_location
        full_path = os.path.join(cls.CONFIG.IO_FILE_ROOT or '', sub_location)
        return full_path if cls.validate(full_path) else None

    @classmethod
    def validate(cls, url: str) -> bool:
        return os.path.isfile(url)

    @classmethod
    def all(cls, root: Optional[str] = None, **kwargs) -> Set[str]:
        root = root or cls.CONFIG.IO_FILE_ROOT or ''
        return {
            os.path.join(dp or '', f or '')
            for dp, dn, filenames in os.walk(root)
            for f in filenames
        }

    def available(self) -> Dict[UrlTypes, ObjectSvc]:
        return {
            UrlTypes.FILE_IO: self,
        }

    def download(self,
                 local_file: BinaryIO,
                 start: Optional[int] = None,
                 end: Optional[int] = None) -> BinaryIO:
        try:
            with open(self.url, 'rb') as this_io:
                shutil.copyfileobj(this_io, local_file)
        except TypeError:
            with open(self.url, 'r') as this_io:
                shutil.copyfileobj(this_io, local_file)
        return local_file

    def download_as_bytes(self,
                          start: Optional[int] = None,
                          end: Optional[int] = None) -> bytes:
        with open(self.url, 'rb') as f:
            content = f.read()
        return content

    def replace(self, replacement: ObjectSvc):
        with open(self.url, 'wb') as dst_file:
            replacement.download(dst_file)

    def delete(self):
        os.remove(self.url)

    @property
    def user_url(self) -> str:
        raise NotImplementedError

    def user_upload_url(self, **kwargs) -> str:
        raise NotImplementedError

    def _get_info(self):
        self._info_cache = {
            'name': os.path.basename(self.url),
            'size': str(os.path.getsize(self.url)),
            'modified': str(os.path.getmtime(self.url)),
            'created': str(os.path.getctime(self.url)),
            'path': self.url,
            'url': self.url
        }

    @property
    def info(self) -> Dict[str, str]:
        if not self._info_cache:
            self._get_info()
        return self._info_cache

    def refresh_info(self) -> Dict[str, str]:
        self._get_info()
        return self._info_cache


class ObjectIngestSvcIO(ObjectIngestSvc):
    """A service to ingest files into file like IO."""

    def __init__(self, base_svc: Optional[Type[ObjectSvc]] = None):
        """Create an instance of to ingest files."""
        super().__init__(base_svc or ObjectSvcIO)

    def ingest(self, ext_file: ObjectSvc, **kwargs) -> ObjectSvc:
        return self.base_svc(self._ingest(ext_file, **kwargs), **kwargs)

    def _ingest(self, ext_file: ObjectSvc, **kwargs) -> str:
        available = ext_file.available()
        if UrlTypes.LOCAL_MEMORY in available:
            return self._ingest_local_memory(available[UrlTypes.LOCAL_MEMORY], **kwargs)
        else:
            raise NotImplementedError("No way to ingest the provided file")

    def _ingest_local_memory(self, file_svc: ObjectSvc, **kwargs) -> str:
        try:
            src_filename = kwargs.pop('filename', None) or file_svc.info.get('name',
                                                                             uuid.uuid4().hex)
        except NotImplementedError:
            src_filename = uuid.uuid4().hex

        dst_path = self.generate_internal_name(src_filename, **kwargs)
        file_mode = kwargs.get('file_mode', 'wb')

        with open(dst_path, file_mode) as dst_file:
            dst_file.write(file_svc.url)
        return dst_path

    def generate_internal_name(self,
                               filename: Optional[str] = None,
                               **kwargs) -> str:
        return os.path.join(self.base_svc.CONFIG.IO_FILE_ROOT or '', filename or '')


class ObjectSvcInMemory(ObjectSvc):
    """A utility service to enable creating new files from memory."""

    @classmethod
    def exists(cls, sub_location: str, root: Optional[str] = None) -> Optional[str]:
        logging.error("You'll need to give the ObjectSvcInMemory the memory "
                      "content when you crate it")
        return None

    @classmethod
    def validate(cls, url: str) -> bool:
        logging.error("The ObjectSvcInMemory must be manually selected")
        return False

    @classmethod
    def all(cls, root: Optional[str] = None, **kwargs) -> Set[str]:
        return set()

    def available(self) -> Dict[UrlTypes, ObjectSvc]:
        return {
            UrlTypes.LOCAL_MEMORY: self,
        }

    def download(self,
                 local_file: BinaryIO,
                 start: Optional[int] = None,
                 end: Optional[int] = None) -> BinaryIO:
        # TODO: Deal with this syntax check error
        #  "Expected type 'bytes' (matched generic type 'AnyStr'), got 'str' instead"
        local_file.write(self.url.encode())
        return local_file

    def download_as_bytes(self,
                          start: Optional[int] = None,
                          end: Optional[int] = None) -> bytes:
        try:
            return self.url.encode()
        except AttributeError:
            return self.url

    @property
    def user_url(self) -> str:
        raise NotImplementedError

    def replace(self, replacement: ObjectSvc):
        self.url = replacement.download_as_bytes()

    def delete(self):
        del self.url
        gc.collect()
        self.url = ''

    @property
    def info(self) -> Dict[str, str]:
        raise NotImplementedError

    def refresh_info(self) -> Dict[str, str]:
        raise NotImplementedError


# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Storage Object Services backed by Google Cloud Storage."""

import uuid
from typing import Dict, BinaryIO, Type, Optional, Set, Union

import requests

from jax.cs.storage.exceptions import ApiAuthError
from jax.cs.storage.object.abstract import ObjectSvc, ObjectIngestSvc, UrlTypes
from . import utils as gcs


class ObjectSvcGCS(ObjectSvc):
    """An ObjectSvc implementation that uses Google Cloud Storage as the storage backend."""

    def __init__(self, url: str, **kwargs):
        """Create an instance of ObjectSvcGCS.

        This initializer performs pre-processing on the url of the GCS object to determine
        bucket_name, file_name, and blob_name during class instantiation.
        :param url: The url of location of the object the service will handle
        :param kwargs: keyword arguments to be used when handling the file
        """
        super().__init__(url, **kwargs)
        self.bucket_name, self.file_name, self.blob_name = gcs.split_gs_uri(self.url)
        self.bucket, self.client = gcs.get_bucket_and_client(
            self.bucket_name,
            credential_content=self.CONFIG.GCS_CREDENTIALS_CONTENTS
        )
        self.blob: gcs.Blob = self.bucket.blob(self.blob_name)

        if self.CONFIG.GCS_CHECK_BLOB_EXISTS:
            self.blob.reload()

    @classmethod
    def _client(cls):
        return gcs.get_client(credential_content=cls.CONFIG.GCS_CREDENTIALS_CONTENTS)

    @classmethod
    def validate(cls, url: str) -> bool:
        _client = cls._client() if cls.CONFIG.GCS_CHECK_BLOB_EXISTS else None
        return gcs.validate_url(url,
                                check_exists=cls.CONFIG.GCS_CHECK_BLOB_EXISTS,
                                client=_client)

    @classmethod
    def all(cls, root: Optional[str] = None, **kwargs) -> Set[str]:
        return gcs.all_files(cls.CONFIG.GCS_BUCKET,
                             prefix=root,
                             full_gs_path=kwargs.get('full_gs_path'),
                             client=cls._client())

    @classmethod
    def exists(cls, sub_location: str, root: Optional[str] = None) -> Optional[str]:
        _client = gcs.get_client(credential_content=cls.CONFIG.GCS_CREDENTIALS_CONTENTS)
        sub_location = f"{root}/{sub_location}" if root else sub_location
        existing = gcs.all_files(cls.CONFIG.GCS_BUCKET,
                                 full_gs_path=False,
                                 client=cls._client())
        if sub_location in existing:
            return f"gs://{cls.CONFIG.GCS_BUCKET}/{sub_location}"
        return None

    def available(self) -> Dict[UrlTypes, ObjectSvc]:
        return {
            UrlTypes.GCS: self,
        }

    def download(self,
                 local_file: BinaryIO,
                 start: Optional[int] = None,
                 end: Optional[int] = None) -> BinaryIO:
        return gcs.download_blob_by_uri(self.client, self.url, local_file, start=start, end=end)

    def download_as_bytes(self,
                          start: Optional[int] = None,
                          end: Optional[int] = None) -> bytes:
        return gcs.download_as_bytes(self.client, self.url, start=start, end=end)

    def replace(self, replacement: ObjectSvc):
        available = replacement.available()
        if UrlTypes.GCS in available and isinstance(available[UrlTypes.GCS], ObjectSvcGCS):
            replacement = available[UrlTypes.GCS]
            gcs.move_object_within_gcs(replacement.bucket,
                                       replacement.blob_name,
                                       self.bucket,
                                       self.blob_name)
        elif UrlTypes.LOCAL_MEMORY in available:
            replacement = available[UrlTypes.LOCAL_MEMORY]
            gcs.upload_string_to_bucket_loaction(self.bucket,
                                                 replacement.url,
                                                 replacement.content_type,
                                                 self.blob_name)
        elif UrlTypes.FILE_IO in available:
            replacement = available[UrlTypes.FILE_IO]
            gcs.upload_file_to_bucket_location(
                self.bucket, replacement.url, self.blob_name
            )
        else:
            gcs.upload_string_to_bucket_loaction(self.bucket,
                                                 replacement.download_as_bytes(),
                                                 replacement.content_type,
                                                 self.blob_name)
        self.refresh_info()

    def delete(self):
        gcs.delete_blob(self.blob)

    @property
    def user_url(self) -> str:
        expire_hrs = self.CONFIG.GCS_URL_EXPIRE_HOURS
        try:
            return gcs.generate_signed_url(self.blob, expiration=expire_hrs)
        except AttributeError as e:
            raise ApiAuthError(e) from e

    def user_upload_url(self, **kwargs) -> str:
        expire_hrs = kwargs.get('expiration') or self.CONFIG.GCS_URL_UPLOAD_EXPIRE_HOURS
        content_type = kwargs.get('content_type') or 'application/octet-stream'
        return gcs.generate_signed_upload_url(self.blob,
                                              content_type=content_type,
                                              expiration=expire_hrs)

    @property
    def info(self) -> Dict[str, str]:
        return {
            'name': self.blob.name,
            'size': self.blob.size,
            'path': self.blob.path,
            'bucket': self.blob.bucket.name,
            'metadata': self.blob.metadata,
            'id': self.blob.id,
            'created': self.blob.time_created,
            'url': self.url
        }

    def refresh_info(self) -> Dict[str, str]:
        self.blob.reload()
        return self.info


class ObjectIngestSvcGCS(ObjectIngestSvc):
    """A Google cloud storage backed ObjectIngestSvc.

    See the documentation for `jax.cs.storage.object.abstract.ObjectIngestSvc`.
    """

    def __init__(self, base_svc: Type[ObjectSvc] = None):
        """Create an ingestion service class instance.

        See documentation on `jax.cs.storage.object.abstract.ObjectIngestSvc`.
        :param base_svc: Optional override for the base Google Cloud storage
        backed ObjectSvc implementation.
        """
        super().__init__(base_svc or ObjectSvcGCS)

    def ingest(self, ext_file: ObjectSvc, **kwargs) -> ObjectSvc:
        """Create a StorageObject instance from ingesting the provided object.

        A factory method to build a ObjectSvc representing a file that is
        internal to the system from a ObjectSvc instance representing a file
        external to the system.
        :param ext_file: An instance of an ObjectSvc representing a file to ingest
        :param kwargs: keyword args to be used by the ingestion service
        :return: An instance of a ObjectSvc after representing the file after being moved
        """
        return self.base_svc(self.__ingest(ext_file, **kwargs), **kwargs)

    def _client(self):
        return gcs.get_client(
            credential_content=self.base_svc.CONFIG.GCS_CREDENTIALS_CONTENTS
        )

    @property
    def object_prefix(self):
        return self.base_svc.CONFIG.GCS_PREFIX_DIR

    @property
    def bucket_name(self):
        return self.base_svc.CONFIG.GCS_BUCKET

    def __internal_validate(self, url: str, check_exists=False):
        if check_exists:
            return (
                    self.__url_in_bucket(url)
                    and
                    gcs.validate_url(url, check_exists=True, client=self._client())
            )
        else:
            return url.startswith(self.bucket_name)

    def __url_in_bucket(self, url):
        url.startswith(self.bucket_name)

    def __generate_internal_name(self, *args, **kwargs):
        filename = '_'.join(args)
        _ = kwargs.pop('filename', None)
        return self.generate_internal_name(filename, uuid=uuid.uuid4().hex, **kwargs)

    def generate_internal_name(self,
                               filename: Optional[str] = None,
                               **kwargs) -> str:
        """Format the name of the ingested object.

        Default template method implementation for GCS.
        Override me to customize object naming behavior in GCS.
        :param filename: The base filename prior to ingestion of the file
        :param kwargs: To be used to pull in extra information to be used in the naming process
        :return: The internal name of the file to use
        """
        ret = f"{self.object_prefix}/" if self.object_prefix else ''
        include_uuid = kwargs.get('include_uuid')
        if include_uuid:
            ret += f"{uuid.uuid4().hex}/"

        return f"{ret}{filename}"

    def __ingest(self, ext_file: ObjectSvc, **kwargs) -> str:
        available = ext_file.available()
        if UrlTypes.GCS in available:
            return self._ingest_gcs_url(available[UrlTypes.GCS], **kwargs)
        elif UrlTypes.LOCAL_MEMORY in available:
            return self._ingest_local_memory(available[UrlTypes.LOCAL_MEMORY], **kwargs)
        elif UrlTypes.FILE_IO in available:
            return self._ingest_local_file(available[UrlTypes.FILE_IO], **kwargs)
        elif UrlTypes.R2 in available:
            return self._ingest_r2_url(available[UrlTypes.R2], **kwargs)
        else:
            raise AttributeError

    def _ingest_gcs_url(self, file_svc: ObjectSvc, **kwargs) -> str:
        src_bucket_name, src_filename, src_blob_name = gcs.split_gs_uri(file_svc.url)
        src_bucket, src_client = gcs.get_bucket_and_client(
            src_bucket_name,
            credential_content=self.base_svc.CONFIG.GCS_CREDENTIALS_CONTENTS
        )
        dst_bucket, dst_client = gcs.get_bucket_and_client(
            self.bucket_name,
            credential_content=self.base_svc.CONFIG.GCS_CREDENTIALS_CONTENTS
        )
        # Override the original filename with one speicifed in the kwargs
        src_filename = kwargs.get('filename', src_filename)
        dst_blob_name = self.__generate_internal_name(src_filename, **kwargs)
        if self.__internal_validate(file_svc.url, check_exists=True):
            # Is it already where we would want to put it?
            if src_bucket_name == self.bucket_name and src_blob_name == dst_blob_name:
                return gcs.gs_path_from_blob(dst_bucket.blob(dst_blob_name))
            else:
                return gcs.move_object_within_gcs(src_bucket, src_blob_name,
                                                  dst_bucket, dst_blob_name)
        elif gcs.validate_url(file_svc.url, check_exists=True, client=src_client):
            return gcs.copy_object_within_gcs(src_bucket, src_blob_name,
                                              dst_bucket, dst_blob_name)
        else:
            raise AttributeError("No way to download provided file")

    def _ingest_r2_url(self, file_svc: ObjectSvc, **kwargs) -> str:
        src_filename = kwargs.get('filename') or file_svc.info.get('name', 'r2_file')
        dst_bucket, dst_client = gcs.get_bucket_and_client(
            self.bucket_name,
            credential_content=self.base_svc.CONFIG.GCS_CREDENTIALS_CONTENTS
        )
        dst_blob_name = self.__generate_internal_name(src_filename, **kwargs)

        resp = requests.get(file_svc.user_url)
        return gcs.upload_string_to_bucket_loaction(dst_bucket,
                                                    resp.content,
                                                    resp.headers['content-type'],
                                                    dst_blob_name)

    def _ingest_local_memory(self, file_svc: ObjectSvc, **kwargs):
        src_filename = kwargs.get('filename') or file_svc.info.get('name', 'local_memory_file')
        content_type = kwargs.get('content_type', 'text/plain')
        dst_bucket, dst_client = gcs.get_bucket_and_client(
            self.bucket_name,
            credential_content=self.base_svc.CONFIG.GCS_CREDENTIALS_CONTENTS
        )
        dst_blob_name = self.__generate_internal_name(src_filename, **kwargs)
        return gcs.upload_string_to_bucket_loaction(dst_bucket,
                                                    file_svc.url,
                                                    content_type,
                                                    dst_blob_name)

    def _ingest_local_file(self, file_svc: ObjectSvc, **kwargs) -> str:
        src_filename = kwargs.get('filename') or file_svc.info.get('name', 'local_file')
        dst_bucket, dst_client = gcs.get_bucket_and_client(
            self.bucket_name,
            credential_content=self.base_svc.CONFIG.GCS_CREDENTIALS_CONTENTS
        )
        dst_blob_name = self.__generate_internal_name(src_filename, **kwargs)
        return gcs.upload_file_to_bucket_location(
            dst_bucket, file_svc.url, dst_blob_name
        )



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module containing enumeration classes for the jax.cs.storage package."""

from enum import Enum


class UrlTypes(Enum):
    """An enumeration of url types specific to this package."""

    HTTPS = 'https'
    R2 = 'r2'
    GCS = 'gs'
    FILE_IO = 'local_file'
    LOCAL_MEMORY = 'local_memory'



# Copyright 2022 The Jackson Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The configuration class and default settings for the jax.cs.storage package."""

import base64
from typing import Optional

from pydantic import AnyHttpUrl, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class StorageConfig(BaseSettings):
    """The pydantic configuration class definition for the jax.cs.storage package."""

    LOG_LEVEL: str = 'DEBUG'

    GCS_CREDENTIALS_B64_DECODE: bool = True
    GCS_CREDENTIALS_CONTENTS: Optional[str] = None

    GCS_BUCKET: Optional[str] = None
    GCS_PREFIX_DIR: Optional[str] = None
    GCS_URL_EXPIRE_HOURS: int = 24
    GCS_URL_UPLOAD_EXPIRE_HOURS: int = 1
    GCS_CHECK_BLOB_EXISTS: bool = False

    R2_BASE_URL: AnyHttpUrl = AnyHttpUrl('https://r2.jax.org')
    IO_FILE_ROOT: Optional[str] = '/'

    @model_validator(mode="after")
    def base64_decode_gcs_credentials(self):
        """Decode the GCS credentials if the flag is set."""
        if self.GCS_CREDENTIALS_B64_DECODE and self.GCS_CREDENTIALS_CONTENTS:
            self.GCS_CREDENTIALS_CONTENTS = base64.b64decode(self.GCS_CREDENTIALS_CONTENTS).decode('utf-8')

    model_config = SettingsConfigDict(
        env_prefix='JAX_CS_STORAGE_',
        env_file=".env",
        case_sensitive=True,
        extra="ignore",
    )


default_settings = StorageConfig()
